"""IUCM - The Integrated Urban Complexity Model

A model to increase the population in an efficient way with respect to
transportation energy
"""

from iucm.version import __version__

__author__ = "Philipp Sommer and Roger Cremades"
__maintainer__ = "Philipp Sommer"
__licence__ = "GPLv3"
__email__ = "philipp.sommer@unil.ch"
__copyright__ = "Copyright 2016, Philipp Sommer and Roger Cremades"
